# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from .type_checked_property import TypeCheckedProperty

__all__ = ["TypeCheckedProperty"]
