# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys


DEFAULT_LOG_PATH = "contrast-agent.log"
DEFAULT_SECURITY_LOG_PATH = "security.log"
DEFAULT_SECURITY_LOG_LEVEL = "INFO"
DEFAULT_LOG_LEVEL = "INFO"
LOGGER_NAME = "contrast"

DEFAULT_AGENT_LOGGER_PATH = sys.stderr
